@@ -72,13 +72,19 @@ class FocusedWave(AcousticField):
             )
 
             # 5. Calculate maximum possible delay
-            max_distance = np.sqrt(
-                max((self.params['Xrange'][0] - self.focal_line)**2,
-                    (self.params['Xrange'][1] - self.focal_line)**2) +
-                self.params['Foc']**2
+            # Calculer max_distance uniquement pour les éléments actifs
+            active_element_positions = element_positions[active_elements]
+            max_distance = np.max(
+                np.sqrt(
+                    (active_element_positions - self.focal_line)**2 +
+                    self.params['Foc']**2
+                )
             )
+
             max_delay_samples = int(np.ceil(max_distance / self.params['c0'] / self.kgrid.dt))
 
+            print(f"max delay (samples): {max_delay_samples}")
+
             # 6. Initialize output array
             delayed_signals = np.zeros((total_grid_points, len(self.burst) + max_delay_samples))
 
@@ -166,7 +172,7 @@ class FocusedWave(AcousticField):
                 current_position += element_width_grid_points
 
             delayed_signals = self._apply_delay(dx=dx) if factorT != 1 else self.delayedSignal
-            source.p = self.params['voltage'] * self.params['sensitivity'] * delayed_signals[active_grid_indices, :]
+            source.p = float(self.params['voltage']) * float(self.params['sensitivity']) * delayed_signals[active_grid_indices, :]
 
         elif source.p_mask.ndim == 3:
             # --- 3D ---
@@ -178,7 +184,7 @@ class FocusedWave(AcousticField):
                 source.p_mask[x_pos, center_index_y, 0] = 1
 
             delayed_signals = self._apply_delay()
-            source.p = self.params['voltage'] * self.params['sensitivity'] * delayed_signals.T
+            source.p = float(self.params['voltage']) * float(self.params['sensitivity']) * delayed_signals.T
 
     def _save2D_HDR_IMG(self, filePath):
         """
