@@ -7,50 +7,116 @@ import numpy as np
 import os
 from tqdm import trange
 
+# def _MLEM_GPU_basic(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
+#     """
+#     Maximum Likelihood Expectation-Maximization (MLEM) for GPU.
+#     """
+#     try:
+#         device = torch.device(f"cuda:{config.select_best_gpu()}")
+#         A_matrix_torch = torch.tensor(SMatrix, dtype=torch.float32).to(device)
+#         y_torch = torch.tensor(y, dtype=torch.float32).to(device)
+#         T, Z, X, N = SMatrix.shape
+#         theta_0 = torch.ones((Z, X), dtype=torch.float32, device=device)
+
+#         A_flat = A_matrix_torch.permute(0, 3, 1, 2).reshape(T * N, Z * X)
+#         y_flat = y_torch.reshape(-1)
+#         matrix_theta_torch = [theta_0]
+#         normalization_factor = A_matrix_torch.sum(dim=(0, 3))
+#         normalization_factor_flat = normalization_factor.reshape(-1)
+
+#         description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM ---- {'WITH' if withTumor else 'WITHOUT'} TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
+
+#         for _ in trange(numIterations, desc=description):
+#             theta_p = matrix_theta_torch[-1]
+#             theta_p_flat = theta_p.reshape(-1)
+#             q_flat = A_flat @ theta_p_flat
+#             e_flat = y_flat / (q_flat + torch.finfo(torch.float32).tiny)
+#             c_flat = A_flat.T @ e_flat
+#             theta_p_plus_1_flat = (theta_p_flat / (normalization_factor_flat + torch.finfo(torch.float32).tiny)) * c_flat
+#             matrix_theta_torch.append(theta_p_plus_1_flat.reshape(Z, X))
+
+#         print("MLEM completed successfully on single GPU, freeing memory.")
+#         del A_matrix_torch, y_torch, A_flat, y_flat, theta_0, normalization_factor, normalization_factor_flat
+#         torch.cuda.empty_cache()
+
+#         if not isSavingEachIteration:
+#             return matrix_theta_torch[-1].cpu().numpy()
+#         else:
+#             return [theta.cpu().numpy() for theta in matrix_theta_torch]
+
+#     except Exception as e:
+#         print("Error in single GPU MLEM:", type(e).__name__, ":", e)
+#         print("Cleaning up resources due to error...")
+#         del A_matrix_torch, y_torch, A_flat, y_flat, theta_0, normalization_factor, normalization_factor_flat
+#         torch.cuda.empty_cache()
+#         return None
+
 def _MLEM_GPU_basic(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
     """
-    Maximum Likelihood Expectation-Maximization (MLEM) for GPU.
+    Optimized MLEM using GPU — fully flattened variables, minimal memory.
     """
     try:
         device = torch.device(f"cuda:{config.select_best_gpu()}")
-        A_matrix_torch = torch.tensor(SMatrix, dtype=torch.float32).to(device)
-        y_torch = torch.tensor(y, dtype=torch.float32).to(device)
+        eps = torch.finfo(torch.float32).eps
+
         T, Z, X, N = SMatrix.shape
-        theta_0 = torch.ones((Z, X), dtype=torch.float32, device=device)
+        ZX = Z * X
+        TN = T * N
 
-        A_flat = A_matrix_torch.permute(0, 3, 1, 2).reshape(T * N, Z * X)
-        y_flat = y_torch.reshape(-1)
-        matrix_theta_torch = [theta_0]
-        normalization_factor = A_matrix_torch.sum(dim=(0, 3))
-        normalization_factor_flat = normalization_factor.reshape(-1)
+        # Conversion directe en A_flat
+        A_flat = (
+            torch.from_numpy(SMatrix)
+            .to(device=device, dtype=torch.float32)
+            .permute(0, 3, 1, 2)     # (T, N, Z, X)
+            .contiguous()
+            .reshape(TN, ZX)
+        )
 
-        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM ---- {'WITH' if withTumor else 'WITHOUT'} TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
+        # Données observées
+        y_flat = torch.from_numpy(y).to(device=device, dtype=torch.float32).reshape(-1)
 
-        for _ in trange(numIterations, desc=description):
-            theta_p = matrix_theta_torch[-1]
-            theta_p_flat = theta_p.reshape(-1)
-            q_flat = A_flat @ theta_p_flat
-            e_flat = y_flat / (q_flat + torch.finfo(torch.float32).tiny)
-            c_flat = A_flat.T @ e_flat
-            theta_p_plus_1_flat = (theta_p_flat / (normalization_factor_flat + torch.finfo(torch.float32).tiny)) * c_flat
-            matrix_theta_torch.append(theta_p_plus_1_flat.reshape(Z, X))
+        # Initialisation
+        theta_flat = torch.ones(ZX, dtype=torch.float32, device=device)
 
-        print("MLEM completed successfully on single GPU, freeing memory.")
-        del A_matrix_torch, y_torch, A_flat, y_flat, theta_0, normalization_factor, normalization_factor_flat
+        # Normalisation directement depuis SMatrix
+        norm_factor_flat = (
+            torch.from_numpy(SMatrix)
+            .to(device=device, dtype=torch.float32)
+            .sum(dim=(0, 3))        # (Z, X)
+            .reshape(-1)
+        )
+
+        # Affichage
+        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM ---- {'WITH' if withTumor else 'WITHOUT'} TUMOR ---- GPU {torch.cuda.current_device()}"
+
+        if isSavingEachIteration:
+            matrix_theta_torch = [theta_flat.reshape(Z, X).clone()]
+
+        with torch.no_grad():
+            for _ in trange(numIterations, desc=description):
+                q_flat = A_flat @ theta_flat
+                e_flat = y_flat / (q_flat + eps)
+                c_flat = A_flat.T @ e_flat
+                theta_flat = (theta_flat / (norm_factor_flat + eps)) * c_flat
+
+                if isSavingEachIteration:
+                    matrix_theta_torch.append(theta_flat.reshape(Z, X).clone())
+
+        # Libération mémoire
+        del A_flat, y_flat, norm_factor_flat
         torch.cuda.empty_cache()
 
-        if not isSavingEachIteration:
-            return matrix_theta_torch[-1].cpu().numpy()
+        if isSavingEachIteration:
+            return [theta.reshape(Z, X).cpu().numpy() for theta in matrix_theta_torch]
         else:
-            return [theta.cpu().numpy() for theta in matrix_theta_torch]
+            return theta_flat.reshape(Z, X).cpu().numpy()
 
     except Exception as e:
-        print("Error in single GPU MLEM:", type(e).__name__, ":", e)
-        print("Cleaning up resources due to error...")
-        del A_matrix_torch, y_torch, A_flat, y_flat, theta_0, normalization_factor, normalization_factor_flat
+        print("Error in MLEM:", type(e).__name__, ":", e)
         torch.cuda.empty_cache()
         return None
 
+
 def _MLEM_CPU_basic(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
     try:
         q_p = np.zeros((SMatrix.shape[0], SMatrix.shape[3]))
@@ -85,35 +151,34 @@ def _MLEM_CPU_basic(SMatrix, y, numIterations, isSavingEachIteration, withTumor)
         print("Error in basic CPU MLEM:", type(e).__name__, ":", e)
         return None
 
-def _MLEM_CPU_multi(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
-    try:
-        numba.set_num_threads(os.cpu_count())
-        q_p = np.zeros((SMatrix.shape[0], SMatrix.shape[3]))
-        c_p = np.zeros((SMatrix.shape[1], SMatrix.shape[2]))
-        theta_p_0 = np.ones((SMatrix.shape[1], SMatrix.shape[2]))
-        matrix_theta = [theta_p_0]
-        normalization_factor = np.sum(SMatrix, axis=(0, 3))
+    def _MLEM_CPU_multi(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
+        try:
+            numba.set_num_threads(os.cpu_count())
+            q_p = np.zeros((SMatrix.shape[0], SMatrix.shape[3]))
+            c_p = np.zeros((SMatrix.shape[1], SMatrix.shape[2]))
+            theta_p_0 = np.ones((SMatrix.shape[1], SMatrix.shape[2]))
+            matrix_theta = [theta_p_0]
+            normalization_factor = np.sum(SMatrix, axis=(0, 3))
 
-        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM ---- {'WITH' if withTumor else 'WITHOUT'} TUMOR ---- processing on multithread CPU ({numba.config.NUMBA_DEFAULT_NUM_THREADS} threads)----"
+            description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM ---- {'WITH' if withTumor else 'WITHOUT'} TUMOR ---- processing on multithread CPU ({numba.config.NUMBA_DEFAULT_NUM_THREADS} threads)----"
 
-        for _ in trange(numIterations, desc=description):
-            theta_p = matrix_theta[-1]
-            _forward_projection(SMatrix, theta_p, q_p)
-            e_p = y / (q_p + 1e-8)
-            _backward_projection(SMatrix, e_p, c_p)
-            theta_p_plus_1 = theta_p / (normalization_factor + 1e-8) * c_p
-            matrix_theta.append(theta_p_plus_1)
+            for _ in trange(numIterations, desc=description):
+                theta_p = matrix_theta[-1]
+                _forward_projection(SMatrix, theta_p, q_p)
+                e_p = y / (q_p + 1e-8)
+                _backward_projection(SMatrix, e_p, c_p)
+                theta_p_plus_1 = theta_p / (normalization_factor + 1e-8) * c_p
+                matrix_theta.append(theta_p_plus_1)
 
-        if not isSavingEachIteration:
-            return matrix_theta[-1]
-        else:
-            return matrix_theta
-
-    except Exception as e:
-        print("Error in multi-threaded CPU MLEM:", type(e).__name__, ":", e)
-        return None
-    
+            if not isSavingEachIteration:
+                return matrix_theta[-1]
+            else:
+                return matrix_theta
 
+        except Exception as e:
+            print("Error in multi-threaded CPU MLEM:", type(e).__name__, ":", e)
+            return None
+        
 def _MLEM_CPU_opti(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
     try:
         T, Z, X, N = SMatrix.shape
